# Copyright (c) 2018 ADLINK Technology Inc.
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors: Angelo Corsaro, ADLINK Technology Inc. - Yaks API refactoring

from yaks.value import Value
from yaks.encoding import Encoding
import zenoh


class Admin(object):
    '''
    The Administration helper class.
    '''

    PREFIX = '@'

    def __init__(self, ws):
        self.ws = ws
        self.local = ''.join('{:02x}'.format(x) for x in
                             ws.rt.info()[zenoh.Z_INFO_PEER_PID_KEY])

    def add_backend(self, beid, properties, yaks=None):
        '''
        Add a backend in the specified Yaks.

        :param beid: the Id of the backend.
        :param propertiers: some configuration for the backend.
        :param yaks: the UUID of the Yaks instance. If ``None``, the local
            Yaks instance.
        '''
        if(yaks is None):
            yaks = self.local
        path = '/{}/{}/plugins/yaks/backend/{}'.format(
            Admin.PREFIX, yaks, beid)
        value = Value(properties, encoding=Encoding.PROPERTY)
        return self.ws.put(path, value)

    def get_backends(self, yaks=None):
        '''
        Get all the backends from the specified Yaks.

        :param yaks: the UUID of the Yaks instance. If ``None``, the local
            Yaks instance.
        '''
        if(yaks is None):
            yaks = self.local
        s = '/{}/{}/plugins/yaks/backend/*'.format(
            Admin.PREFIX, yaks)
        entries = self.ws.get(s)
        return list(map(
            lambda e: (e.get_path().split('/')[-1],
                       e.get_value().value), entries))

    def get_backend(self, beid, yaks=None):
        '''
        Get backend's properties from the specified Yaks.

        :param beid: the Id of the backend.
        :param yaks: the UUID of the Yaks instance. If ``None``, the local
            Yaks instance.
        '''
        if(yaks is None):
            yaks = self.local
        s = '/{}/{}/plugins/yaks/backend/{}'.format(
            Admin.PREFIX, yaks, beid)
        entries = self.ws.get(s)
        if len(entries) > 0:
            return entries[0].get_value().value
        return None

    def remove_backend(self, beid, yaks=None):
        '''
        Remove a backend from the specified Yaks.

        :param beid: the Id of the backend.
        :param yaks: the UUID of the Yaks instance. If ``None``, the local
            Yaks instance.
        '''
        if(yaks is None):
            yaks = self.local
        path = '/{}/{}/plugins/yaks/backend/{}'.format(
            Admin.PREFIX, yaks, beid)
        return self.ws.remove(path)

    def add_storage(self, stid, properties, beid=None, yaks=None):
        '''
        Adds a storage in the specified Yaks backend.

        :param stid: the Id of the storage.
        :param propertiers: some configuration for the storage.
        :param beid: the Id of the backend. If ``None``, a backend is
            automatically selected.
        :param yaks: the UUID of the Yaks instance. If ``None``, the local
            Yaks instance.
        '''
        if(yaks is None):
            yaks = self.local
        if not beid:
            beid = 'auto'
        p = '/{}/{}/plugins/yaks/backend/{}/storage/{}'.format(
            Admin.PREFIX, yaks, beid, stid)
        v = Value(properties, encoding=Encoding.PROPERTY)
        return self.ws.put(p, v)

    def get_storages(self, beid=None, yaks=None):
        '''
        Adds a storage in the specified Yaks.

        :param beid: the Id of the backend. If ``None``, all backends.
        :param yaks: the UUID of the Yaks instance. If ``None``, the local
            Yaks instance.
        '''
        if(yaks is None):
            yaks = self.local
        if not beid:
            beid = '*'
        s = '/{}/{}/plugins/yaks/backend/{}/storage/*'.format(
            Admin.PREFIX, yaks, beid)
        entries = self.ws.get(s)
        return list(map(
            lambda e: (e.get_path().split('/')[-1],
                       e.get_value().value), entries))

    def get_storage(self, stid, yaks=None):
        '''
        Get storage's properties from the specified Yaks.

        :param stid: the Id of the storage.
        :param yaks: the UUID of the Yaks instance. If ``None``, the local
            Yaks instance.
        '''
        if(yaks is None):
            yaks = self.local
        s = '/{}/{}/plugins/yaks/backend/*/storage/{}'.format(
            Admin.PREFIX, yaks, stid)
        entries = self.ws.get(s)
        if len(entries) > 0:
            return entries[0].get_value().value
        return None

    def remove_storage(self, stid, yaks=None):
        '''
        Remove a backend from the specified Yaks.

        :param stid: the Id of the storage.
        :param yaks: the UUID of the Yaks instance. If ``None``, the local
            Yaks instance.
        '''
        if(yaks is None):
            yaks = self.local
        s = '/{}/{}/plugins/yaks/backend/*/storage/{}'.format(
            Admin.PREFIX, yaks, stid)
        entries = self.ws.get(s)
        if len(entries) > 0:
            p = entries[0].get_path()
            return self.ws.remove(p)
        return False



# Copyright (c) 2018 ADLINK Technology Inc.
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors: Gabriele Baldoni, ADLINK Technology Inc. - Yaks API

import zenoh


class Entry(object):
    def __init__(self, path, value, timestamp):
        self.path = path
        self.value = value
        self.timestamp = timestamp

    def __hash__(self):
        # As timestamp is unique per entry, only hash the timestamp.
        return hash(self.timestamp)

    def __eq__(self, other):
        # As timestamp is unique per entry, only compare timestamps.
        return self.timestamp.__eq__(other.timestamp)

    def __lt__(self, other):
        # order entries according to timestamps
        return self.timestamp.__lt__(other.timestamp)

    def get_path(self):
        return self.path

    def get_value(self):
        return self.value

    def get_timestamp(self):
        return self.timestamp



# Copyright (c) 2018 ADLINK Technology Inc.
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors: Gabriele Baldoni, ADLINK Technology Inc. - Yaks API

import re
import json
import zenoh
from enum import Enum
from yaks.exceptions import ValidationError
from yaks.encoding import Encoding


class ChangeKind(Enum):
    PUT = zenoh.Z_PUT
    UPDATE = zenoh.Z_UPDATE
    REMOVE = zenoh.Z_REMOVE


class Value(object):
    def __init__(self, value, encoding=Encoding.RAW, raw_format=""):
        if encoding is None:
            encoding = Encoding.RAW
        if encoding > Encoding.MAX:
            raise ValueError('Encoding not supported')
        if encoding == Encoding.PROTOBUF:
            raise ValueError('PROTOBUF Encoding not implemented')
        self.encoding = encoding
        if self.encoding == Encoding.JSON:
            if not (isinstance(value, dict) or isinstance(value, str)):
                raise ValidationError("Value is not a valid JSON")
            self.value = json.dumps(value)
        elif self.encoding == Encoding.RAW and isinstance(value, str):
            self.value = value.encode()
        else:
            self.value = value
        self.raw_format = raw_format

    def as_z_payload(self):
        if self.encoding == Encoding.RAW:
            return self.value
        if self.encoding == Encoding.PROPERTY:
            s = ';'.join(map('='.join, map(list, self.value.items())))
            return s.encode()
        return self.value.encode()

    def get_encoding(self):
        return self.encoding

    def get_value(self):
        if self.encoding is Encoding.JSON:
            return json.loads(self.value)
        return self.value

    def __eq__(self, second_value):
        if isinstance(second_value, self.__class__):
            return self.value == second_value.value
        return False

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return self.__str__()

    @staticmethod
    def from_z_resource(buf, info):
        encoding = Encoding.from_z_encoding(info.encoding)
        data = None
        if(encoding == Encoding.RAW):
            data = bytearray(buf)
        else:
            data = buf.decode()
        return Value(data, encoding)


class Change(object):
    kind_map = {
            zenoh.Z_PUT: ChangeKind.PUT,
            zenoh.Z_UPDATE: ChangeKind.UPDATE,
            zenoh.Z_REMOVE: ChangeKind.REMOVE
    }

    def __init__(self, path, kind, time, value=None):
        self.path = path
        self.time = time
        self.value = value
        if kind is None:
            self.kind = ChangeKind.PUT
        elif isinstance(kind, ChangeKind):
            self.kind = kind
        else:
            self.kind = Change.kind_map[kind]

    def get_path(self):
        return self.path

    def get_kind(self):
        return self.kind

    def get_time(self):
        return self.time

    def get_value(self):
        return self.value

    def __str__(self):
        return 'Path: {} Kind: {} Time: {} Value: {}'.format(
            self.path,
            self.kind,
            self.time,
            self.value)

    def __repr__(self):
        return self.__str__()



# Copyright (c) 2018 ADLINK Technology Inc.
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors: Angelo Corsaro, ADLINK Technology Inc. - Yaks API refactoring

from queue import Queue
from yaks.encoding import Encoding, TranscodingFallback
from yaks.path import Path
from yaks.selector import Selector
from yaks.value import Value, Change
from yaks.entry import Entry
import zenoh
from zenoh import *


class Workspace(object):
    '''

    A Workspace to operate on Yaks.

    '''

    def __init__(self, runtime, path, executor=None):
        self.rt = runtime
        self.path = Path.to_path(path)
        self.evals = []
        self.executor = executor

    def __to_absolute(self, path):
        if path.startswith('/'):
            return path
        else:
            return self.path.to_string() + '/' + path

    def put(self, path, value):
        '''

        Put a path/value into Yaks.

        :param path: the Path. Can be absolute or relative to the workspace.
        :param value: the value.

        '''

        self.rt.write_data(
            self.__to_absolute(path),
            value.as_z_payload(),
            Encoding.to_z_encoding(value.get_encoding()),
            zenoh.Z_PUT)
        return True

    def update(self, path, value):
        '''

        Update a path/value into Yaks.

        :param path: the Path. Can be absolute or relative to the workspace.
        :param value: a delta to be applied on the existing value.

        '''

        raise NotImplementedError("Update not yet implemented ...")

    def __isSelectorForSeries(self, selector):
        props = selector.get_properties()
        if props is None:
            return True
        for p in props.split(";"):
            if(p.startswith("starttime") or p.startswith("stoptime")):
                return True
        return False

    def get(self, selector, encoding=Encoding.RAW,
                fallback=TranscodingFallback.KEEP):
        '''

        Get a selection of path/value from Yaks.

        :param selector: the selector expressing the selection.
        :returns: a list of entry.

        '''

        q = Queue()

        def callback(reply_value):
            q.put(reply_value)

        def contains_key(kvs, key):
            for (k, _) in kvs:
                if(k == key):
                    return True
            return False

        selector = Selector.to_selector(self.__to_absolute(selector))

        self.rt.query(
            selector.get_path(),
            selector.get_optional_part(),
            callback)
        resultsMap = {}
        reply = q.get()
        while(reply.kind != zenoh.Z_REPLY_FINAL):
            if(reply.kind == zenoh.Z_STORAGE_DATA
               or reply.kind == zenoh.Z_EVAL_DATA):
                entry = Entry(reply.rname,
                              Value.from_z_resource(reply.data, reply.info),
                              reply.info.tstamp)
                if reply.rname not in resultsMap:
                    resultsMap[reply.rname] = set()
                resultsMap[reply.rname].add(entry)
            reply = q.get()
        q.task_done()

        results = []
        if(self.__isSelectorForSeries(selector)):
            # return all entries
            for path, entrySet in resultsMap.items():
                for entry in sorted(entrySet):
                    results.append(entry)
        else:
            # return only the latest entry for each path
            for path, entrySet in resultsMap.items():
                entries = sorted(entrySet)
                results.append(entries[-1])

        return results

    def remove(self, path):
        '''

        Remove a path/value from Yaks.

        :param path: the Path to be removed.
            Can be absolute or relative to the workspace.

        '''

        self.rt.write_data(
            self.__to_absolute(path),
            "".encode(),
            Encoding.Z_RAW_ENC,
            zenoh.Z_REMOVE)
        return True

    def subscribe(self, selector, listener):
        '''

        Subscribe to a selection of path/value from Yaks.

        :param selector: the selector expressing the selection.
        :param listener: the Listener that will be called for each change of
            a path/value matching the selection.
        :returns: a subscription id.

        '''

        selector = self.__to_absolute(selector)
        if(listener is not None):
            def callback(rname, data, info):
                if self.executor is None:
                    listener([Change(
                        rname,
                        info.kind,
                        info.tstamp.time if info.tstamp is not None else None,
                        Value.from_z_resource(data, info))])
                else:
                    self.executor.submit(listener, [Change(
                        rname,
                        info.kind,
                        info.tstamp.time if info.tstamp is not None else None,
                        Value.from_z_resource(data, info))])
            return self.rt.declare_subscriber(
                selector,
                zenoh.SubscriberMode.push(),
                callback)
        else:
            def callback(rname, data, info):
                pass
            return self.rt.declare_subscriber(
                selector,
                zenoh.SubscriberMode.push(),
                callback)

    def unsubscribe(self, subscription_id):
        '''

        Unregisters a previous subscription.

        :param subscription_id: the subscription id to unregister.

        '''

        self.rt.undeclare_subscriber(subscription_id)
        return True

    def register_eval(self, path, callback):
        '''

        Registers an evaluation function under the provided path.

        :param path: the Path where the function can be triggered using
            :func:`~yaks.workspace.Workspace.get`.
        :param callback: the evaluation function.

        '''

        def query_handler(path_selector, content_selector, send_replies):
            def query_handler_p(path_selector, content_selector, send_replies):
                args = Selector.dict_from_properties(
                    Selector("{}?{}".format(path_selector, content_selector)))
                value = callback(path_selector, args)
                info = z_data_info_t()
                info.flags = 0x60
                info.encoding = Encoding.to_z_encoding(value.get_encoding())
                info.kind = Z_PUT
                send_replies([(path_selector, (value.as_z_payload(), info))])
            if self.executor is None:
                query_handler_p(path_selector,
                                content_selector,
                                send_replies)
            else:
                self.executor.submit(query_handler_p,
                                     path_selector,
                                     content_selector,
                                     send_replies)

        path = self.__to_absolute(path)
        zeval = self.rt.declare_eval(path,
                                     query_handler)

        self.evals.append((path, zeval))

    def unregister_eval(self, path):
        '''

        Unregister a previously registered evaluation function.

        :param path: the path where the function has been registered.

        '''

        path = self.__to_absolute(path)
        for (evalpath, zeval) in self.evals:
            if evalpath == path:
                self.rt.undeclare_eval(zeval)
        self.evals = [(evalpath, zeval) for (evalpath, zeval) in
                      self.evals if not evalpath == path]
        return True



# Copyright (c) 2018 ADLINK Technology Inc.
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors: Angelo Corsaro, ADLINK Technology Inc. - Yaks API refactoring

from yaks.workspace import Workspace
from yaks.admin import *
import threading
from zenoh import Zenoh, Z_INFO_PEER_PID_KEY


class Yaks(object):
    '''

    The Yaks client API.

    '''

    ZENOH_DEFAULT_PORT = 7447

    def __init__(self, rt):
        self.rt = rt

    @staticmethod
    def login(locator, properties=None):
        '''

        Establish a session with the Yaks instance reachable via provided
        Zenoh locator. If the provided locator is ``None``, :func:`login`
        will perform some dynamic discovery and try to establish the session
        automatically. When not ``None``, the locator must have the format:
        ``tcp/<ip>:<port>``.

        :param locator: a Zenoh locator or ``None``.
        :param properties: the Properties to be used for this session
            (e.g. "user", "password", ...). Can be ``None``.
        :returns: a Yaks object.

        '''
        zprops = {} if properties is None else {
            zenoh.Z_USER_KEY if k == "user" else zenoh.Z_PASSWORD_KEY: val
            for k, val in properties.items()
            if k == "user" or key == "password"}

        return Yaks(Zenoh.open(locator, zprops))

    def workspace(self, path, executor=None):
        '''

        Creates a :class:`~yaks.workspace.Workspace` using the
        provided path. All relative Selector or Path used with this
        :class:`~yaks.workspace.Workspace` will be relative to
        this path.

        :param path: the Workspace's path.
        :param executor: an executor of type
            :py:class:`concurrent.futures.Executor` or ``None``.
            If not ``None``, all subscription listeners and eval callbacks are
            executed by the provided executor. This is useful when listeners
            and/or callbacks need to perform long operations or need to call
            operations like :func:`~yaks.workspace.Workspace.get`.
        :returns: a :class:`~yaks.workspace.Workspace`.

        '''
        return Workspace(self.rt, path, executor)

    def logout(self):
        '''

        Terminates this session.

        '''
        self.rt.close()

    def admin(self):
        '''

        Creates an admin workspace that provides helper operations to
        administer Yaks.

        '''
        return Admin(self.workspace(
            '/{}/{}'.format(
                Admin.PREFIX,
                ''.join('{:02x}'.format(x) for x in
                        self.rt.info()[Z_INFO_PEER_PID_KEY]))))
